from datetime import timedelta
from typing import Dict, List

import pandas as pd
from tqdm import tqdm

from . import data_inputs, evaluate_EWRs, summarise_results
from mdba_gauge_getter import gauge_getter as gg

def categorise_gauges(gauges):
    '''Seperate gauges into level, flow, or both'''
    menindee_gauges, weirpool_gauges = data_inputs.get_level_gauges()
    multi_gauges = data_inputs.get_multi_gauges('gauges')
    simultaneous_gauges = data_inputs.get_simultaneous_gauges('gauges')
    
    level_gauges = []
    flow_gauges = []
    # Loop through once to get the special gauges:
    for gauge in gauges:
        if gauge in multi_gauges.keys():
            flow_gauges.append(gauge)
            flow_gauges.append(multi_gauges[gauge])
        if gauge in simultaneous_gauges:
            flow_gauges.append(gauge)
            flow_gauges.append(simultaneous_gauges[gauge])
        if gauge in menindee_gauges:
            level_gauges.append(gauge)
        if gauge in weirpool_gauges.keys(): # need level and flow gauges
            flow_gauges.append(gauge)
            level_gauges.append(weirpool_gauges[gauge])
    # Then loop through again and allocate remaining gauges to the flow category
    for gauge in gauges:
        if ((gauge not in level_gauges) and (gauge not in flow_gauges)):
            # Otherwise, assume its a flow gauge and add
            flow_gauges.append(gauge)
        
        unique_flow_gauges = list(set(flow_gauges))
        unique_level_gauges = list(set(level_gauges))
            
    return unique_flow_gauges, unique_level_gauges

def observed_handler(gauges, dates, allowance, climate):
    '''ingests a list of gauges and user defined parameters
    pulls gauge data using relevant states API, calcualtes and analyses EWRs
    returns dictionary of raw data results and result summary
    '''
    
    # Classify gauges:
    flow_gauges, level_gauges = categorise_gauges(gauges)
    # Call state API for flow and level gauge data, then combine to single dataframe
    
    flows = gg.gauge_pull(flow_gauges, start_time_user = dates['start_date'], end_time_user = dates['end_date'], var = 'F')
    levels = gg.gauge_pull(level_gauges, start_time_user = dates['start_date'], end_time_user = dates['end_date'], var = 'L')
    # Clean observed data:
    df_F = observed_cleaner(flows, dates)
    df_L = observed_cleaner(levels, dates)
    # Calculate EWRs
    detailed_results = {}
    gauge_results = {}
    gauge_events = {}
    detailed_events = {}
    all_locations = df_F.columns.to_list() + df_L.columns.to_list()
    for gauge in all_locations:
        gauge_results[gauge], gauge_events[gauge] = evaluate_EWRs.calc_sorter(df_F, df_L, gauge, allowance, climate)
        
    detailed_results['observed'] = gauge_results
    detailed_events['observed'] = gauge_events
    # Summarise the results:
    summary_results = summarise_results.summarise(detailed_results, detailed_events)

    return detailed_results, summary_results


def remove_data_with_bad_QC(input_dataframe, qc_codes):
    '''Takes in a dataframe of flow and a list of bad qc codes, removes the poor quality data from 
    the timeseries, returns this dataframe'''
    for qc in qc_codes:
        input_dataframe.loc[input_dataframe.QUALITYCODE == qc, 'VALUE'] = None
        
    return input_dataframe

def one_gauge_per_column(input_dataframe, gauge_iter):
    '''Takes in a dataframe and the name of a gauge, extracts this one location to a new dataframe, 
    cleans this and returns the dataframe with only the selected gauge data'''
    
    is_in = input_dataframe['SITEID']== gauge_iter
    single_df = input_dataframe[is_in]
    single_df = single_df.drop(['DATASOURCEID','SITEID','SUBJECTID','QUALITYCODE','DATETIME'],
                               axis = 1)
    single_df = single_df.set_index('Date')
    single_df = single_df.rename(columns={'VALUE': str(gauge_iter)})
    
    return single_df

def observed_cleaner(input_df, dates):
    '''Takes in raw dataframe consolidated from state websites, removes poor quality data.
    returns a dataframe with a date index and one flow column per gauge location.'''
    
    
    df_index = pd.date_range(dates['start_date'],dates['end_date']-timedelta(days=1),freq='d')
    gauge_data_df = pd.DataFrame()
    gauge_data_df['Date'] = df_index
    gauge_data_df['Date'] = pd.to_datetime(gauge_data_df['Date'], format = '%Y-%m-%d')
    gauge_data_df = gauge_data_df.set_index('Date')

    input_df["VALUE"] = pd.to_numeric(input_df["VALUE"])#, downcast="float")
    
    
    input_df['Date'] = pd.to_datetime(input_df['DATETIME'], format = '%Y-%m-%d')
    # Check with states for more codes:
    bad_data_codes = data_inputs.get_bad_QA_codes()
    input_df = remove_data_with_bad_QC(input_df, bad_data_codes)
    
    site_list = set(input_df['SITEID'])
    
    for gauge in site_list:
        # Seperate out to one gauge per column and add this to the gauge_data_df made above:
        single_gauge_df = one_gauge_per_column(input_df, gauge)
        gauge_data_df = pd.merge(gauge_data_df, single_gauge_df, left_index=True, right_index=True, how="outer")

    # Drop the non unique values:
    gauge_data_df = gauge_data_df[~gauge_data_df.index.duplicated(keep='first')]
    return gauge_data_df


class ObservedHandler:
    
    def __init__(self, gauges:List, dates:Dict , allowance:Dict, climate:str):
        self.gauges = gauges
        self.dates = dates
        self.allowance = allowance
        self.climate = climate
        self.EWR_TABLE = None
        self.yearly_events = None
        self.pu_ewr_statistics = None
        self.summary_results = None

    def process_gauges(self):
        '''ingests a list of gauges and user defined parameters
        pulls gauge data using relevant states API, calculates and analyses EWRs
        returns dictionary of raw data results and result summary
        '''
        
        # Classify gauges:
        flow_gauges, level_gauges = categorise_gauges(self.gauges)
        # Call state API for flow and level gauge data, then combine to single dataframe
        
        flows = gg.gauge_pull(flow_gauges, start_time_user = self.dates['start_date'], end_time_user = self.dates['end_date'], var = 'F')
        levels = gg.gauge_pull(level_gauges, start_time_user = self.dates['start_date'], end_time_user = self.dates['end_date'], var = 'L')
        # Clean observed data:
        df_F = observed_cleaner(flows, self.dates)
        df_L = observed_cleaner(levels, self.dates)
        # Calculate EWRs
        detailed_results = {}
        gauge_results = {}
        gauge_events = {}
        detailed_events = {}
        all_locations = df_F.columns.to_list() + df_L.columns.to_list()
        for gauge in all_locations:
            gauge_results[gauge], gauge_events[gauge] = evaluate_EWRs.calc_sorter(df_F, df_L, gauge, self.allowance, self.climate)
            
        detailed_results['observed'] = gauge_results
        detailed_events['observed'] = gauge_events
        
        self.pu_ewr_statistics = detailed_results
        self.yearly_events = detailed_events


    def get_all_events(self)-> pd.DataFrame:

        if not self.yearly_events:
            self.process_gauges()
        
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events = summarise_results.process_all_events_results(events_to_process)
        return all_events

    def get_yearly_ewr_results(self)-> pd.DataFrame:

        if not self.pu_ewr_statistics:
            self.process_gauges()

        to_process = summarise_results.pu_dfs_to_process(self.pu_ewr_statistics)
        yearly_ewr_results = summarise_results.process_df_results(to_process)
        return yearly_ewr_results

    def get_ewr_results(self) -> pd.DataFrame:
        
        if not self.pu_ewr_statistics:
            self.process_gauges()

        return summarise_results.summarise(self.pu_ewr_statistics , self.yearly_events)


import io
import requests
import pandas as pd
import numpy as np
from pathlib import Path

BASE_PATH = Path(__file__).resolve().parent

# Importing the climate cat data - to be replaced by RAS data once available:
   
def get_climate_cats(climate_file):
    '''Uses standard climate categorisation unless user selects the 10,000 year climate sequence,
    in which case this is used'''
    
    if climate_file == 'Standard - 1911 to 2018 climate categorisation':
        climate_cats = pd.read_csv( BASE_PATH / 'climate_data/climate_cats.csv', index_col = 0)
        
    elif climate_file  == 'NSW 10,000 year climate sequence':
        climate_cats = pd.read_csv(BASE_PATH / 'climate_data/climate_cats_10000year.csv', index_col = 0)
        
    return climate_cats

def get_EWR_table(my_url = 'https://az3mdbastg001.blob.core.windows.net/mdba-public-data/NSWEWR_LIVE_DEV.csv'):
    
    ''' Loads ewr table from blob storage, seperates out the readable ewrs from the 
    ewrs with 'see notes' exceptions, those with no threshold, and those with undefined names,
    does some cleaning, including swapping out '?' in the frequency column with 0'''
    
    proxies={} # Populate with your proxy settings
    s = requests.get(my_url, proxies=proxies).text
    df = pd.read_csv(io.StringIO(s),
                     usecols=['PlanningUnitID', 'PlanningUnitName',  'CompliancePoint/Node', 'gauge', 'code', 'start month',
                              'end month', 'frequency', 'events per year', 'duration', 'min event', 'flow threshold min', 'flow threshold max',
                              'max inter-event', 'within event gap tolerance', 'weirpool gauge', 'flow level volume', 'level threshold min',
                              'level threshold max', 'volume threshold', 'drawdown rate'],
                     dtype='str'
                    )

    df = df.replace('?','')
    df = df.fillna('')

    # removing the 'See notes'
    okay_EWRs = df.loc[(df["start month"] != 'See note') & (df["end month"] != 'See note')]
    see_notes = df.loc[(df["start month"] == 'See note') & (df["end month"] == 'See note')]

    # Filtering those with no flow/level/volume thresholds
    noThresh_df = okay_EWRs.loc[(okay_EWRs["flow threshold min"] == '') & (okay_EWRs["flow threshold max"] == '') &\
                             (okay_EWRs["volume threshold"] == '') &\
                             (okay_EWRs["level threshold min"] == '') & (okay_EWRs["level threshold max"] == '')]
    okay_EWRs = okay_EWRs.loc[(okay_EWRs["flow threshold min"] != '') | (okay_EWRs["flow threshold max"] != '') |\
                        (okay_EWRs["volume threshold"] != '') |\
                        (okay_EWRs["level threshold min"] != '') | (okay_EWRs["level threshold max"] != '')]

    # Filtering those with no durations
    okay_EWRs = okay_EWRs.loc[(okay_EWRs["duration"] != '')]
    no_duration = df.loc[(df["duration"] == '')]

    # FIltering DSF EWRs
    condDSF = df['code'].str.startswith('DSF')
    DSF_ewrs = df[condDSF]
    condDSF_inv = ~(okay_EWRs['code'].str.startswith('DSF'))
    okay_EWRs = okay_EWRs[condDSF_inv]

    # Combine the unuseable EWRs into a dataframe and drop dups:
    bad_EWRs = pd.concat([see_notes, noThresh_df, no_duration, DSF_ewrs], axis=0)
    bad_EWRs = bad_EWRs.drop_duplicates()

    # Changing the flow and level max threshold to a high value when there is none available:
    okay_EWRs['flow threshold max'].replace({'':'1000000'}, inplace = True)
    okay_EWRs['level threshold max'].replace({'':'1000000'}, inplace = True)
    
    return okay_EWRs, bad_EWRs


def map_gauge_to_catchment(my_url = 'https://az3mdbastg001.blob.core.windows.net/mdba-public-data/NSWEWR_LIVE_DEV.csv'):
    ''' Allocates all the locations in the ewr table with catchments, as indicated by the
    first three numbers for each gauge '''
    
    lower_darling_gauges = ['425054', '425010', '425011', '425052', '425013', '425056', '425007', 
                            '425057', '425005', '425050', '425048', '425019', '425014', '425023', 
                            '425012', '425044', '425049', '425001', '425022', '42510037', '42510036',
                            '425034', '425046', '425020', ]
    
    EWR_table, bad_EWRs =  get_EWR_table(my_url)
    
    gauge_number = EWR_table['gauge'].values
    gauge_name = EWR_table['CompliancePoint/Node'].values
    
    gauge_to_name = dict()
    for iteration, value in enumerate(gauge_number):
        if type(value) == str:
            gauge_to_name[value] = gauge_name[iteration]

    gauge_to_catchment = dict()
    namoi_catchment = dict()
    gwydir_catchment = dict()
    macquarie_catchment = dict()
    lachlan_catchment = dict()
    murray_catchment = dict()
    lower_darling_catchment = dict()
    barwon_darling_catchment = dict()
    murrumbidgee_catchment = dict()
    border_rivers_catchment = dict()
    moonie_catchment = dict()
    condamine_balonne = dict()
    warrego_catchment = dict()
    paroo_catchment = dict()
    
    for k, v in gauge_to_name.items():
        if k.startswith('419'):
            namoi_catchment.update({k: v})
        elif k.startswith('418'):
            gwydir_catchment.update({k: v})
        elif (k.startswith('421') or k.startswith('420')):
            macquarie_catchment.update({k: v})
        elif k.startswith('412'):
            lachlan_catchment.update({k: v})
        elif (k.startswith('401') or k.startswith('409') or k.startswith('426') or k.startswith('414')):
            murray_catchment.update({k: v})
        elif k.startswith('425'):
            if k in lower_darling_gauges:
                lower_darling_catchment.update({k: v})   
            else:
                barwon_darling_catchment.update({k: v}) 
        elif k.startswith('410'):
            murrumbidgee_catchment.update({k: v})
        elif k.startswith('416'):
            border_rivers_catchment.update({k: v})
        elif k.startswith('417'):
            moonie_catchment.update({k: v})
        elif k.startswith('422'):
            condamine_balonne.update({k: v})
        elif k.startswith('423'):
            warrego_catchment.update({k: v})
        elif k.startswith('424'):
            paroo_catchment.update({k: v})            
                
    gauge_to_catchment.update({'Namoi': namoi_catchment, 
                               'Gwydir': gwydir_catchment,
                               'Macquarie-Castlereagh': macquarie_catchment, 
                               'Lachlan': lachlan_catchment,
                               'Lower Darling': lower_darling_catchment, 
                               'Barwon-Darling': barwon_darling_catchment,
                               'Murray': murray_catchment,
                               'Murrumbidgee': murrumbidgee_catchment,
                               'Border Rivers': border_rivers_catchment,
                               'Moonie' : moonie_catchment,
                               'Condamine-Balonne': condamine_balonne,
                               'Warrego': warrego_catchment,
                               'Paroo': paroo_catchment
                              })
    return gauge_to_catchment

def get_MDBA_codes():
    '''
    Load MDBA model metadata file containing model nodes
    and gauges they correspond to
    '''
    metadata = pd.read_csv( BASE_PATH / 'model_metadata/SiteID_MDBA.csv', engine = 'python', dtype=str, encoding='windows-1252')
#     metadata = metadata.where(pd.notnull(metadata), None)

    return metadata
  
def get_NSW_codes():
    '''
    Load NSW model metadata file containing model nodes
    and gauges they correspond to
    '''
    metadata = pd.read_csv( BASE_PATH / 'model_metadata/SiteID_NSW.csv', engine = 'python', dtype=str)
    
    return metadata

def gauge_to_catchment(input_gauge):
    '''
    Takes in a gauge, maps it to the catchment
    returns the catchment
    '''
    catchments_gauges = map_gauge_to_catchment()
    for catchment in catchments_gauges:
        if input_gauge in catchments_gauges[catchment]:
            return catchment
    
def wy_to_climate(water_years, catchment, climate_file):
    '''
    water_years = a daily water year value array
    catchment = the catchment that the gauge is in
    climate file = which climate data to use
    
    The function assigns a climate categorisation for every day, depending on the water year and catchment in the climate file
    '''
    # Get the climate categorisation:
    climate_cats = get_climate_cats(climate_file)
    
    # Get the unique years covered in the flow dataframe, and how many days are in each year:
    unique_years, count_years = np.unique(water_years, return_counts=True)
    
    # Get the years covered by the climate cats, and filter them to those of interest (using min and max from flow dataframe)
    climateCatchment = climate_cats[catchment]
    climateFiltered = climateCatchment[(climateCatchment.index>=min(unique_years)) & (climateCatchment.index<=max(unique_years))].values
    # Repeating the climate result for that year over the total days in each year 
    def mapper(climate, count):
        return np.repeat(climate, count)

    climateDailyYear = list(map(mapper, climateFiltered, count_years))
    climateDaily = np.concatenate(climateDailyYear)
    
    return climateDaily


def get_level_gauges():
    '''Returning level gauges with EWRs'''
    
    menindeeGauges = ['425020', '425022', '425023']
    
    weirpoolGauges = {'414203': '414209', 
                      '425010': '4260501', 
                      '4260507': '4260508',
                      '4260505': '4260506'}
    
    return menindeeGauges, weirpoolGauges


def get_multi_gauges(dataType):
    '''
    Call function to return a dictionary of multi gauges.
    Multi gauges are for EWRs that require the flow of two gauges to be added together
    '''
    
    gauges = {'PU_0000130': {'421090': '421088', '421088': '421090'},
              'PU_0000131': {'421090': '421088', '421088': '421090'},
              'PU_0000132': {'421090': '421088', '421088': '421090'},
              'PU_0000133': {'421090': '421088', '421088': '421090'}
             }
    returnData = {}
    if dataType == 'all':
        returnData = gauges
    if dataType == 'gauges':
        for i in gauges:
            returnData = {**returnData, **gauges[i]}
    
    return returnData

def get_simultaneous_gauges(dataType):
    '''
    Call function to return a dictionary of simultaneous gauges.
    Simultaneous gauges are for EWRs that need to be met simultaneously with EWRs at another location
    '''
    
    gauges = {'PU_0000131': {'421090': '421022', '421022': '421090'},
              'PU_0000132': {'421090': '421022', '421022': '421090'},
              'PU_0000133': {'421090': '421022', '421022': '421090'}
             }
    returnData = {}
    if dataType == 'all':
        returnData = gauges
    if dataType == 'gauges':
        for i in gauges:
            returnData = {**returnData, **gauges[i]}
        
    return returnData

def get_complex_calcs():
    '''
    Returns a dictionary of the complex EWRs, and the type of analysis that needs to be undertaken
    These EWRs cannot be calculated using the standard suite of functions
    '''
    complexCalcs = {'409025': {'OB2_S': 'flowDurPostReq', 'OB2_P': 'flowDurPostReq',
                              'OB3_S': 'flowDurOutsideReq', 'OB3_P': 'flowDurOutsideReq'}}
    
    return complexCalcs


def get_gauges(category):
    '''
    Gathers a list of either all gauges that have EWRs associated with them,
    a list of all flow type gauges that have EWRs associated with them,
    or a list of all level type gauges that have EWRs associated with them
    '''
    EWR_table, bad_EWRs = get_EWR_table()
    menindee_gauges, wp_gauges = get_level_gauges()
    wp_gauges = list(wp_gauges.values())
    
    multi_gauges = get_multi_gauges('gauges')
    simul_gauges = get_simultaneous_gauges('gauges')
    multi_gauges = list(multi_gauges.values())
    simul_gauges = list(simul_gauges.values())
    if category == 'all gauges':
        return set(EWR_table['gauge'].to_list() + menindee_gauges + wp_gauges + multi_gauges + simul_gauges)
    elif category == 'flow gauges':
        return set(EWR_table['gauge'].to_list() + multi_gauges + simul_gauges)
    elif category == 'level gauges':
        return set(menindee_gauges + wp_gauges)
    else:
        raise ValueError('''No gauge category sent to the "get_gauges" function''')

def get_EWR_components(category):
    '''
    Ingests an EWR type and a gauge, returns the components required to analyse for
    this type of EWR. Each code stands for a component.    
    '''

    if category == 'flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'GP', 'EPY', 'MIE']
    elif category == 'low flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR',  'ME', 'EPY', 'DURVD', 'MIE']
    elif category == 'cease to flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'EPY', 'DURVD', 'MIE']
    elif category == 'cumulative':
        pull =  ['SM', 'EM', 'MINV', 'DUR', 'ME', 'EPY', 'MINF', 'MIE']
    elif category == 'level':
        pull = ['SM', 'EM', 'MINL', 'MAXL', 'DUR', 'ME', 'EPY', 'MD', 'MIE']
    elif category == 'weirpool-raising':
        pull=['SM', 'EM', 'MINF', 'MAXF', 'MINL', 'DUR', 'ME',  'MD', 'EPY','WPG', 'MIE']
    elif category == 'weirpool-falling':
        pull=['SM', 'EM', 'MINF', 'MAXF', 'MAXL', 'DUR', 'ME',  'MD', 'EPY','WPG', 'MIE']
    elif category == 'nest-level':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME',  'MD', 'EPY', 'WPG', 'MIE']
    elif category == 'nest-percent':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME',  'MD', 'EPY', 'MIE']
    elif category == 'multi-gauge-flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME',  'GP', 'EPY', 'MG', 'MIE']
    elif category == 'multi-gauge-low flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'EPY', 'DURVD', 'MG', 'MIE']
    elif category == 'multi-gauge-cease to flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'EPY', 'DURVD', 'MG', 'MIE']
    elif category == 'multi-gauge-cumulative':
        pull =  ['SM', 'EM', 'MINV', 'DUR', 'ME', 'EPY', 'MINF', 'MG', 'MIE']
    elif category == 'simul-gauge-flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'GP', 'EPY', 'DURVD', 'SG', 'MIE']
    elif category == 'simul-gauge-low flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'GP', 'EPY', 'DURVD', 'SG', 'MIE']
    elif category == 'simul-gauge-cease to flow':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME', 'GP', 'EPY', 'DURVD', 'SG', 'MIE']
    elif category == 'complex':
        pull = ['SM', 'EM', 'MINF', 'MAXF', 'DUR', 'ME',  'GP', 'EPY', 'MIE']  
    return pull
    
def get_bad_QA_codes():
    '''These codes are NSW specific'''
    return [151, 152, 153, 155, 180, 201, 202, 204, 205, 207, 223, 255]
    

def additional_nest_pull(EWR_info, gauge, EWR, allowance):
    '''Additional EWR information not yet included in the database has been hard coded'''
    if ((gauge == '409025') and ('NestS1' in EWR)):
        EWR_info['trigger_day'] = 15
        EWR_info['trigger_month'] = 9
    elif ((gauge == '409207') and ('NestS1' in EWR)):
        EWR_info['trigger_day'] = 1
        EWR_info['trigger_month'] = 10
        EWR_info['start_month'] = 10
        EWR_info['start_day'] = None
    elif 'NestS1' in EWR:
        EWR_info['trigger_day'] = None
        EWR_info['trigger_month'] = None
        
    return EWR_info

def analysis():
    '''Returns a list of types of analysis to be shown in the summary table'''
    
    return ['Event years','Frequency','Target frequency','Achievement count', 'Achievements per year', 'Event count','Events per year',
            'Event length','Threshold days','Inter-event exceedence count', 'Max inter event period (years)', 'No data days',
            'Total days']

def weirpool_type(EWR):
    '''Returns the type of Weirpool EWR. Currently only WP2 EWRs are classified as weirpool raisings'''
    if EWR == 'WP2':
        weirpool_type = 'raising'
    else:
        weirpool_type = 'falling'

    return weirpool_type

def convert_max_interevent(unique_water_years, water_years, EWR_info):
    '''Max interevent is saved in the database as years, we want to convert it to days.'''
    new_max_interevent = 365 * EWR_info['max_inter-event']
    
    return new_max_interevent

def get_planning_unit_info():
    '''Run this function to get the planning unit MDBA ID and equivilent planning unit name as specified in the LTWP'''
    EWR_table, bad_EWRs = get_EWR_table()
        
    planningUnits = EWR_table.groupby(['PlanningUnitID', 'PlanningUnitName']).size().reset_index().drop([0], axis=1) 
    
    return planningUnits


from ipywidgets import Button, Text, VBox, HBox, Layout, Dropdown, Checkbox, \
    DatePicker, Select, SelectMultiple, Tab, BoundedFloatText, Label, Output, interactive
from IPython.display import display
import traitlets
import itertools
import pandas as pd
import copy
from tkinter import Tk, filedialog
from datetime import datetime, date
import xlsxwriter

from . import data_inputs, scenario_handling, observed_handling

#-------------------------------------------------------------------------------------------------#
#-----------------------------------Tab 1 - Observed flow widgets---------------------------------#
#-------------------------------------------------------------------------------------------------#

# Input tab widgets--------------------------------------------------------------------------------
# Label widgets:
defaults_label = {'style': {'description_width':'initial'}}
gauge_inputs_title = Label(value="Configure inputs for observed EWRs", **defaults_label)
date_selection = Label(value="1. Select date range of interest", **defaults_label)
cs_selection = Label(value="2. Select the catchment (1) and sites (1+)", **defaults_label)
allowance_header_o = Label(value="3. Enter an allowance", **defaults_label)
run_header_o = Label(value="4. Enter output file name and run", **defaults_label)

# Date selection widgets:
start_date = DatePicker(description='Start date:', value = date(2014,1,1), disabled=False)
end_date = DatePicker(description='End date', value = date(2020,1,1), disabled=False)

# Catchment and site selection widgets:
defaults_sites = {'layout': Layout(width="100%"), 'style': {'description_width':'initial'}}


def change_catchment(*args):
    '''Updates the sites displayed based on the user selection of the catchment'''
    catchments_gauges = data_inputs.map_gauge_to_catchment()
    sites.options=list(catchments_gauges[catchment.value].values())
    
catchments_gauges = data_inputs.map_gauge_to_catchment()

catchment = Select(options=list(catchments_gauges.keys()), 
                           description='Catchment: (select one)',
                           rows=len(catchments_gauges.keys()), **defaults_sites)

sites = SelectMultiple(options=catchments_gauges[catchment.value].values(),
                               description='Sites: (shift to select multiple)',
                               rows=len(catchments_gauges[catchment.value].values()),
                               **defaults_sites)
catchment.observe(change_catchment, 'value')

# Apply allowance to EWR indicators:
defaults = {'value': 0, 'min': 0, 'max': 20, 'step': 0.1, 'disabled': False,
            'style': {'description_width':'initial'}}
min_threshold_allowance_o = BoundedFloatText(description='Allowance applied to min threshold (%)', **defaults)
max_threshold_allowance_o = BoundedFloatText(description='Allowance applied to max threshold (%)', **defaults)
duration_allowance_o = BoundedFloatText(description='Allowance applied to min duration (%)', **defaults)
drawdown_allowance_o = BoundedFloatText(description='Allowance applied to max drawdown rate (%)', **defaults)

# Output file name:
file_name_o = Text(value='Observed flow EWRs', placeholder='Enter file name', 
                           description='Output file name:', disabled=False, 
                           style= {'description_width':'initial'})
# Results tab--------------------------------------------------------------------------------------
def get_gauges_to_pull():
    '''Convert sites selected to gauge numbers'''
    catchments_gauges = data_inputs.map_gauge_to_catchment()
    gauges = []
    for gauge, name in catchments_gauges[catchment.value].items():
        if name in sites.value:
            gauges.append(gauge)
    return gauges
    
def on_gauge_button_clicked(b):
    '''Run the realtime program'''
    with gauge_run_output:
        b.style.button_color='lightgreen'
        tab_gauge.selected_index = 1
        global raw_data_o, results_summary_o
        raw_data_o, results_summary_o = None, None
        # Get gauge list:
        gauges = get_gauges_to_pull()

        # Retrieve and convert dates:
#         dates = {'start_date': str(start_date.value).replace('-',''), 
#                  'end_date': str(end_date.value).replace('-','')}
        dates = {'start_date': start_date.value, 
                 'end_date': end_date.value}
        # Get allowances:
        MINT = (100 - min_threshold_allowance_s.value)/100
        MAXT = (100 + max_threshold_allowance_s.value)/100
        DUR = (100 - duration_allowance_s.value)/100
        DRAW = (100 - drawdown_allowance_s.value)/100
        allow ={'minThreshold': MINT, 'maxThreshold': MAXT, 'duration': DUR, 'drawdown': DRAW}
        # hardcode in the climate cats for the observed flows:
        raw_data_o, results_summary_o = observed_handling.observed_handler(gauges, dates, allow,
                                                                           'Standard - 1911 to 2018 climate categorisation') 
        display(results_summary_o.style)

def gauge_output_button_clicked(b):
    '''Output the realtime analysis to excel'''
    
    with gauge_output:
        b.style.button_color='lightgreen'
        global results_summary_o
        global raw_data_o
        
        realtime_fileName = file_name_o.value # Gettng the user file name        
        writer = pd.ExcelWriter('Output_files/' + realtime_fileName + '.xlsx', engine='xlsxwriter')
        results_summary_o.to_excel(writer, sheet_name='Gauge_data_summary') 
        PU_items = data_inputs.get_planning_unit_info()
        PU_items.to_excel(writer, sheet_name='Planning unit metadata')
        
        get_index = results_summary_o.reset_index().set_index(['gauge', 'planning unit']).index.unique()
        for locPU in get_index:
            temp_df = pd.DataFrame()
            if temp_df.empty == True:
                temp_df = raw_data_o['observed'][locPU[0]][locPU[1]].copy(deep=True)
                temp_df.columns = pd.MultiIndex.from_product([[str('observed')],temp_df.columns])
            else:
                df_to_add = pd.DataFrame()
                df_to_add = raw_data_o['observed'][locPU[0]][locPU[1]].copy(deep=True)
                df_to_add.columns = \
                pd.MultiIndex.from_product([[str('observed')],df_to_add.columns])        
                temp_df = pd.concat([temp_df, df_to_add], axis = 1)
                
            PU_code = PU_items['PlanningUnitID'].loc[PU_items[PU_items['PlanningUnitName'] == locPU[1]].index[0]]
            temp_df.to_excel(writer, sheet_name=str(PU_code))                
                
        writer.save()
#Run the analysis (from input tab):
gauge_run_button = Button(description="Run gauge program")
gauge_run_output = Output(layout={'border': '1px solid black'})
gauge_run_button.on_click(on_gauge_button_clicked)
# Output results tab-------------------------------------------------------------------------------
default_output = {'style': {'description_width':'initial'}}
gauge_output_button = Button(description='Output to Excel', **default_output)
gauge_output = Output(layout={'border': '1px solid black', **default_output})
gauge_output_button.on_click(gauge_output_button_clicked)
#-------------------------------------------------------------------------------------------------#
#-----------------------------------Tab 2 - Scenario testing widgets------------------------------#
#-------------------------------------------------------------------------------------------------#

# Input tab----------------------------------------------------------------------------------------
#Labels:
defaults_label = {'style': {'description_width':'initial'}}

model_inputs_title = Label(value="Configure inputs for scenario testing", **defaults_label)
model_selection = Label(value="1. Upload model files", **defaults_label)
model_format_header = Label(value="2. Select format of timeseries data", **defaults_label)
climate_format_header = Label(value="3. Select climate sequence to load", **defaults_label)
allowance_header_s = Label(value="4. Enter an allowance", **defaults_label)
run_header_s = Label(value="5. Enter output file name and run", **defaults_label)
fileLocationHeader = Label(value="Provide file path and names:", **defaults_label)
# Local file upload:
class SelectFilesButton(Button):
    """A file widget that leverages tkinter.filedialog."""

    def __init__(self):
        super(SelectFilesButton, self).__init__()
        # Add the selected_files trait
        self.add_traits(files=traitlets.traitlets.List())
        # Create the button.
        self.description = "Select Files"
        self.icon = "square-o"
        self.style.button_color = "orange"
        # Set on click behavior.
        self.on_click(self.select_files)
        
    @staticmethod
    def select_files(b):
        global file_names
        global select_benchmark
        """Generate instance of tkinter.filedialog.

        Parameters
        ----------
        b : obj:
            An instance of ipywidgets.widgets.Button 
        """
        try:
            # Create Tk root
            root = Tk()
            # Hide the main window
            root.withdraw()
            # Raise the root to the top of all windows.
            root.call('wm', 'attributes', '.', '-topmost', True)
            # List of selected fileswill be set to b.value
            b.files = filedialog.askopenfilename(multiple=True)
            b.description = "Files Selected"
            b.icon = "check-square-o"
            b.style.button_color = "lightgreen"
        except:
            print('no file selected')
            pass
load_model_files = SelectFilesButton()        
# Remote file upload:
box_defaults = {'value':'', 'placeholder':'Enter/path/to/file', 'disabled': False, 
                'style': {'description_width':'initial'}}
path_box_1 = Text(description='Scenario file one name:', **box_defaults)
path_box_2 = Text(description='Scenario file two name:', **box_defaults)
path_box_3 = Text(description='Scenario file three name:', **box_defaults)
path_box_4 = Text(description='Scenario file four name:', **box_defaults)
path_box_5 = Text(description='Scenario file five name:', **box_defaults)

#Allowance widgets:
default = {'value': 0, 'min': 0, 'max': 20, 'step': 0.1, 'disabled': False,
            'style': {'description_width':'initial'}}
min_threshold_allowance_s = BoundedFloatText(description='Allowance applied to min threshold (%)', **defaults)
max_threshold_allowance_s = BoundedFloatText(description='Allowance applied to max threshold (%)', **defaults)
duration_allowance_s = BoundedFloatText(description='Allowance applied to min duration (%)', **defaults)
drawdown_allowance_s = BoundedFloatText(description='Allowance applied to max drawdown rate (%)', **defaults)

# Model format widget:
model_format_type = Dropdown(
    options=[('Bigmod - MDBA'), ('IQQM - NSW 10,000 years'), ('Source - NSW (res.csv)')],
    value='Bigmod - MDBA',
    description='',
    style= {'description_width':'initial'})
# Climate file widget:
climate_file = Dropdown(
    options=[('Standard - 1911 to 2018 climate categorisation'), ('NSW 10,000 year climate sequence')],
    value='Standard - 1911 to 2018 climate categorisation',
    description='',
    style= {'description_width':'initial'})
# Output file name:
file_name_s = Text(
    value='Scenario_test',
    placeholder='Enter file name',
    description='Output file name:',
    disabled=False,
    style= {'description_width':'initial'})
# Results display tab------------------------------------------------------------------------------

def catchment_checker(catchment):
    '''Pass catchment name, returns list of gauges in this catchment'''
    catchments_gauges = data_inputs.map_gauge_to_catchment()
    return list(catchments_gauges[catchment].keys()) 
    
def view(x=''):
    '''Shows either all results, or results restricted to user catchment selection'''
    if x=='All': 
        return display(data_summary_s.style)
    return display((data_summary_s.loc[data_summary_s.index.isin(catchment_checker(x), level='gauge')]).style)

def get_locations_from_scenarios(data_summary):
    '''Ingest a summary of results, look at the locations analysed, 
    return a list of catchments included in the analysis'''
    catchments_gauges = data_inputs.map_gauge_to_catchment()
    location_list = set(data_summary_s.index.get_level_values(0))
    catchments_gauges_subset = copy.deepcopy(catchments_gauges)
    for catch in catchments_gauges.keys():
        for site in catchments_gauges[catch]:
            if site not in location_list:
                del catchments_gauges_subset[catch][site]
            else:
                continue        
    items = ['All']+list(catchments_gauges_subset.keys())   
    
    return items

def get_file_names(loaded_files):
    '''Take in the file location strings from the users loaded files,
    return dictionary with file location string mapped to file name'''
    
    file_locations = {}
    for file in loaded_files:
        full_name = file.split('/')
        name_exclude_extension = full_name[-1].split('.csv')[0]
        file_locations[str(name_exclude_extension)] = file
        
    return file_locations

def on_model_button_clicked(b):
    '''Run the scenario testing program'''
    
    with model_run_output:
        b.style.button_color='lightgreen'
        tab_model.selected_index = 1
        global raw_data_s, data_summary_s, select_benchmark
        # Get file names and their system locations
        if load_model_files.files != []:
            modelFiles = load_model_files.files
        else:
            modelFiles = []
            if path_box_1.value != '':
                PB1 = path_box_1.value.strip()
                modelFiles.append(PB1)
            if path_box_2.value != '':
                PB2 = path_box_2.value.strip()
                modelFiles.append(PB2)
            if path_box_3.value != '':
                PB3 = path_box_3.value.strip()
                modelFiles.append(PB3)
            if path_box_4.value != '':
                PB4 = path_box_4.value.strip()
                modelFiles.append(PB4)
            if path_box_5.value != '':
                PB5 = path_box_5.value.strip()
                modelFiles.append(PB5)
        file_locations = get_file_names(modelFiles)
        # Get tolerance:
        minThreshold_tolerance = (100 - min_threshold_allowance_s.value)/100
        maxThreshold_tolerance = (100 + max_threshold_allowance_s.value)/100
        duration_tolerance = (100 - duration_allowance_s.value)/100
        drawdown_tolerance = (100 - drawdown_allowance_s.value)/100

        allowanceDict ={'minThreshold': minThreshold_tolerance, 'maxThreshold': maxThreshold_tolerance, 
                        'duration': duration_tolerance, 'drawdown': drawdown_tolerance}
        
        # Run the requested analysis on the loaded scenarios:
        raw_data_s, data_summary_s = scenario_handling.scenario_handler(file_locations,
                                                                        model_format_type.value,
                                                                        allowanceDict,
                                                                        climate_file.value)
        items = get_locations_from_scenarios(data_summary_s)              
        w = Select(options=items)
        #show the results of the selected catchment:
        out = interactive(view, x=w) 
        display(out)

def getMetadata():
    '''Run this function to get the metadata from the model run.
    Returns a pandas dataframe with the relevant metadata'''
    todaysDate = str(datetime.today().strftime('%Y-%m-%d')) 

    metadata = pd.DataFrame(columns = ['Name', 'Details'])
    nameList = ['date run', 'database accessed on', 'duration tolerance applied', 'minimum threshold tolerance applied', 'maximum threshold tolerance applied', 
                'maximum drawdown rate tolerance applied']
    metadataList = [todaysDate, todaysDate, duration_allowance_s.value, min_threshold_allowance_s.value, max_threshold_allowance_s .value, drawdown_allowance_s.value]
    metadata['Name'] = nameList
    metadata['Details'] = metadataList
    
    return metadata

def model_output_button_clicked(b):
    '''Output the scenario testing to excel'''
    with model_output:
        b.style.button_color='lightgreen'
        global data_summary_s
        global raw_data_s
        
        # Get the metadata from the model run:
        metadata_df = getMetadata()
        PU_items = data_inputs.get_planning_unit_info()
        model_file_name = file_name_s.value # Gettng the user file name
        
        model_scenario_list = list(data_summary_s.columns.levels[0]) # Get scenarios / Repeated?
        
        writer = pd.ExcelWriter('Output_files/' + model_file_name + '.xlsx', engine='xlsxwriter')
        metadata_df.to_excel(writer, sheet_name='Metadata')
        PU_items.to_excel(writer, sheet_name='Planning unit metadata')
        data_summary_s.to_excel(writer, sheet_name='Data_summary')
        get_index = data_summary_s.reset_index().set_index(['gauge', 'planning unit']).index.unique()
        for locPU in get_index:
            temp_df = pd.DataFrame()
            for model_scenario in model_scenario_list:
                if temp_df.empty == True:
                    temp_df = raw_data_s[model_scenario][locPU[0]][locPU[1]].copy(deep=True)
                    temp_df.columns = pd.MultiIndex.from_product([[str(model_scenario)],temp_df.columns])
                else:
                    df_to_add = pd.DataFrame()
                    df_to_add = raw_data_s[model_scenario][locPU[0]][locPU[1]].copy(deep=True)
                    df_to_add.columns = \
                    pd.MultiIndex.from_product([[str(model_scenario)],df_to_add.columns])
                    temp_df = pd.concat([temp_df, df_to_add], axis = 1)
                PU_code = PU_items['PlanningUnitID'].loc[PU_items[PU_items['PlanningUnitName'] == locPU[1]].index[0]]
                sheet_name = str(locPU[0]) + '-' + str(PU_code)
                temp_df.to_excel(writer, sheet_name=sheet_name)

        writer.save()
#Run analysis button (from input tab):
model_run_button = Button(description="Run program")
model_run_output = Output(layout={'border': '1px solid black'}) 
model_run_button.on_click(on_model_button_clicked)
#Output results tab--------------------------------------------------------------------------------
model_output_button = Button(description="Output to Excel")
model_output = Output(layout={'border': '1px solid black'}, style= {'description_width':'initial'})
model_output_button.on_click(model_output_button_clicked)
#-------------------------------------------------------------------------------------------------#
#--------------------------------------Widgets used by both tabs----------------------------------#
#-------------------------------------------------------------------------------------------------#
justLine = HBox([Label(value="________________________________________________________________________________", 
                       layout=Layout(width="100%"),
                       style = {'description_width':'initial'})])
                       
or_text = HBox([Label(value="~     OR     ~", 
                                      layout=Layout(width="100%"),
                                      style= {'description_width':'initial'})])
#-------------------------------------------------------------------------------------------------#
#-------------------------------------------Aggregate widgets-------------------------------------#
#-------------------------------------------------------------------------------------------------#
gauge_input_widgets = VBox([gauge_inputs_title, justLine, date_selection, HBox([start_date, end_date]), 
                                     justLine, cs_selection, HBox([catchment, sites]),
                                     justLine, allowance_header_o, HBox([min_threshold_allowance_o, max_threshold_allowance_o, 
                                                                         duration_allowance_o, drawdown_allowance_o]),
                                     justLine, run_header_o, HBox([file_name_o, gauge_run_button])])

model_input_widgets= VBox([model_inputs_title, justLine, model_selection, 
                                   VBox([load_model_files, or_text, fileLocationHeader, path_box_1, path_box_2, path_box_3, path_box_4, path_box_5]),
                                   VBox([justLine, model_format_header, model_format_type]),
                                   VBox([justLine, climate_format_header, climate_file]),
                                   justLine, allowance_header_s, HBox([min_threshold_allowance_s, max_threshold_allowance_s, duration_allowance_s, drawdown_allowance_s]),
                                   justLine, run_header_s, HBox([file_name_s, model_run_button])])

model_output_widgets = HBox([model_output_button, model_output])
gauge_output_widgets = HBox([gauge_output_button, gauge_output])
# Then group them together:    
tab_gauge = Tab()
tab_gauge.children = [gauge_input_widgets, gauge_run_output, gauge_output_widgets]
tab_gauge.set_title(0, 'Input')
tab_gauge.set_title(1, 'Results dashboard')
tab_gauge.set_title(2, 'Output results') 
    
tab_model = Tab()
tab_model.children = [model_input_widgets, model_run_output, model_output_widgets]
tab_model.set_title(0, 'Input')
tab_model.set_title(1, 'Results dashboard')
tab_model.set_title(2, 'Output results') 

ewr_dashboard = Tab()
ewr_dashboard.children = [tab_gauge, tab_model]
ewr_dashboard.set_title(0, 'Observed')
ewr_dashboard.set_title(1, 'Scenario testing')


from typing import List, Dict
from itertools import chain
from collections import defaultdict

import pandas as pd

from . import data_inputs
#--------------------------------------------------------------------------------------------------


def get_frequency(events):
    '''Returns the frequency of years they occur in'''
    if events.count() == 0:
        result = 0
    else:
        result = (int(events.sum())/int(events.count()))*100
    return int(round(result, 0))


def get_ewr_columns(ewr:str, cols:List) -> List:
    """Filter the columns of a particular ewr code in a list of 
    column names.

    Args:
        ewr (str): Ewr code
        cols (List): list of columns to search ewr pattern

    Returns:
        List: List of columns that matches the ewr code
    """
    return [c for c in cols if ewr in c]


def get_columns_attributes(cols: List)-> List:
    """Takes a list of columns with the pattern EwrCode_Attribute
    and relates them returning only the Attribute name.

    Args:
        cols (List): DataFrame columns names as a list

    Returns:
        List: List of the column attribute stripped out of the ewr code
    """
    return [c.split("_")[-1] for c  in cols]

def get_ewrs(pu_df: pd.DataFrame)-> List:
    """Take a DataFrame with the location results and by searching its 
    column return a list with the unique name of ewrs on the results.

    Args:
        pu_df (pd.DataFrame): DataFrame with with location stats results

    Returns:
        List: Returns a list os unique ewrs present in the location results
    """
    cols = pu_df.columns.to_list()
    ewrs_set = set(["_".join(c.split("_")[:-1]) for c  in cols])
    return list(ewrs_set)

def pu_dfs_to_process(detailed_results: Dict)-> List[Dict]:
    """Take the detailed results dictionary of the ewr calculation run,
    either observed or scenario and unpack items into a list of items.
    Each item is a dictionary with the following keys.
                { "scenario" : scenario_name,
                  "gauge" : gauge_id,
                  "pu" : pu_name,
                  "pu_df : DataFrame}

    Args:
        detailed_results (Dict): Dictionary with the following structure
        { "scenario_name/or observed": {"gaugeID": {"pu_name": pu_DateFrame}
        
            }

        } 
        It packs in a dictionary all the gauge ewr calculation for the scenario
        or observed dates run.

    Returns:
        List[Dict]: list of dict with the items to be processed
    """
    items_to_process = []
    for scenario in detailed_results:
        for gauge in detailed_results[scenario]:
            for pu in detailed_results[scenario][gauge]:
                item = {}
                item["scenario"] = scenario
                item["gauge"] = gauge
                item["pu"] = pu
                item["pu_df"] = detailed_results[scenario][gauge][pu]
                items_to_process.append(item)
    return items_to_process


def process_df(scenario:str, gauge:str, pu:str, pu_df: pd.DataFrame)-> pd.DataFrame:
    """Process all the pu_dfs into a tidy format

    Args:
        scenario (str): scenario name metadata
        gauge (str): gauge name metadata
        pu (str): planning unit name metadata
        pu_df (pd.DataFrame): DataFrame to be transformed

    Returns:
        pd.DataFrame: DataFrame with all processed pu_dfs into one.
    """
    ewrs = get_ewrs(pu_df)
    returned_dfs = []
    for ewr in ewrs:
        columns_ewr = get_ewr_columns(ewr, pu_df.columns.to_list())
        ewr_df = pu_df[columns_ewr]
        column_attributes = get_columns_attributes(ewr_df.columns.to_list())
        ewr_df.columns = column_attributes
        ewr_df = ewr_df.reset_index().rename(columns={"index":'Year'})
        ewr_df["ewrCode"] = ewr
        ewr_df["scenario"] = scenario
        ewr_df["gauge"] = gauge
        ewr_df["pu"] = pu
        ewr_df["daysBetweenEvents"] = ewr_df["daysBetweenEvents"].agg(len)
        ewr_df = ewr_df.loc[:,~ewr_df.columns.duplicated()]
        returned_dfs.append(ewr_df)
    return pd.concat(returned_dfs, ignore_index=True)


def process_df_results(results_to_process: List[Dict])-> pd.DataFrame:
    """Manage the processing and concatenating the processed dfs into one
    single dataframe with the results of all ewr calculations.

    Args:
        results_to_process (List[Dict]): List with all items to process.

    Returns:
        pd.DataFrame: Single DataFrame with all the ewr results
    """
    returned_dfs = []
    for item in results_to_process:
        transformed_df = process_df(**item)
        returned_dfs.append(transformed_df)
    return pd.concat(returned_dfs, ignore_index=True)

def get_events_to_process(gauge_events: dict)-> List:
    """Take the detailed gauge events results dictionary of the ewr calculation run,
    and unpack items into a list of items.
    Each item is a dictionary with the following keys.
                { "scenario" : scenario_name,
                  "gauge" : gauge_id,
                  "pu" : pu_name,
                  "ewr": ewr_code
                  "ewr_events" : yearly_events_dictionary}

    Args:
        gauge_events (dict): Gauge events captured by the ewr calculations.
        Dictionary with the following structure
        {'observed': {'419001': {'Keepit to Boggabri': {'CF1_a': ({2010: [],
                2011: [],
                2012: [],
                2013: [],
                2014: [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]},)}
                                }
                     }
        }
        It packs in a dictionary all the gauge ewr yearly events and threshold flows.

    Returns:
        List: list of dict with the items to be processed
    """
    items_to_process = []
    for scenario in gauge_events:
        for gauge in gauge_events[scenario]:
            for pu in gauge_events[scenario][gauge]:
                for ewr in gauge_events[scenario][gauge][pu]:
                    item = {}
                    item["scenario"] = scenario
                    item["gauge"] = gauge
                    item["pu"] = pu
                    item["ewr"] = ewr
                    item["ewr_events"],  = gauge_events[scenario][gauge][pu][ewr]
                    items_to_process.append(item)
    return items_to_process


def count_events(yearly_events:dict)-> int:
    """count the events in a collection of years

    Args:
        yearly_events (dict): ewr yearly events dictionary of lists of lists

    Returns:
        int: count of length of all events in the collection of years
    """
    return sum([len(events) for events in yearly_events.values()])


def sum_events(yearly_events:dict)-> int:
    """sum the total event days in a collection of years

    Args:
        yearly_events (dict): ewr yearly events dictionary of lists of lists

    Returns:
        int: count of total days of all events in the collection of years
    """

    flattened_events = [list(chain(*events)) for events in yearly_events.values()]
    return len(list(chain(*flattened_events)))


def process_yearly_events(scenario:str, gauge:str, pu:str, ewr:str, ewr_events: Dict)-> pd.DataFrame:
    """process each item for the gauge and return the statistics in a DataFrame

    Args:
        scenario (str): scenario name metadata
        gauge (str): gauge name metadata
        pu (str): planning unit name metadata
        ewr (str): DataFrame to be transformed
        ewr_events (Dict): Dict with all yearly events list with date and flow/level 

    Returns:
        pd.DataFrame: DataFrame with events statistics
    """
    row_data = defaultdict(list)
    yearly_events = ewr_events
    total_events = count_events(yearly_events)
    total_event_days = sum_events(yearly_events)
    average_event_length = total_event_days/total_events if total_events else 0
    row_data['scenario'].append(scenario)
    row_data['gauge'].append(gauge)
    row_data['pu'].append(pu)
    row_data['ewrCode'].append(ewr)
    row_data['totalEvents'].append(total_events)
    row_data['totalEventDays'].append(total_event_days)
    row_data['averageEventLength'].append(average_event_length)
    
    return pd.DataFrame(row_data)

def process_ewr_events_stats(events_to_process: List[Dict])-> pd.DataFrame:
    """Manage the processing of yearly events and concatenate into a
    single dataframe with the results of all ewr calculations.

    Args:
        events_to_process (List[Dict]): List with all items to process.

    Returns:
        pd.DataFrame: Single DataFrame with all the ewr events stats results
    """
    returned_dfs = []
    for item in events_to_process:
        row_data = process_yearly_events(**item)
        returned_dfs.append(row_data)
    return pd.concat(returned_dfs, ignore_index=True)

def process_all_yearly_events(scenario:str, gauge:str, pu:str, ewr:str, ewr_events: Dict)-> pd.DataFrame():
    """process each item for the gauge and return all events. Each event is a row with a start and end date
    duration and event length

    Args:
        scenario (str): scenario name metadata
        gauge (str): gauge name metadata
        pu (str): planning unit name metadata
        ewr (str): DataFrame to be transformed
        ewr_events (Dict): Dict with all yearly events list with date and flow/level
        

    Returns:
        pd.DataFrame: DataFrame with all events of Pu-ewr-gauge combination
    """
    df_data = defaultdict(list)
    for year in ewr_events:
        for i, ev in enumerate(ewr_events[year]):
            start_date, _ = ev[0]
            end_date, _ = ev[-1]
            df_data["scenario"].append(scenario)
            df_data["gauge"].append(gauge)
            df_data["pu"].append(pu)
            df_data["ewr"].append(ewr)
            df_data["waterYear"].append(year)
            df_data["startDate"].append(start_date )
            df_data["endDate"].append(end_date)
            df_data["eventDuration"].append((end_date - start_date).days + 1)
            df_data["eventLength"].append(len(ev))          
    
    return pd.DataFrame(df_data)

def process_all_events_results(results_to_process: List[Dict])-> pd.DataFrame:
    """Manage the processing of yearly events and concatenate into a
    single dataframe with the results of all ewr calculations.

    Args:
        results_to_process (List[Dict]):List with all items to process.

    Returns:
        pd.DataFrame: Single DataFrame with all the ewr events
    """
    returned_dfs = []
    for item in results_to_process:
        df = process_all_yearly_events(**item)
        returned_dfs.append(df)
    return pd.concat(returned_dfs, ignore_index=True)


def summarise(input_dict:Dict , events:Dict)-> pd.DataFrame:
    """orchestrate the processing of the pu_dfs items and the gauge events and join
    in one summary DataFrame and join with EWR parameters for comparison

    Args:
        input_dict (Dict): DataFrame result by yearly with statistics for the ewr calculations.
        events (Dict): Gauge events captured by the ewr calculations.

    Returns:
        pd.DataFrame: Summary statistics for all ewr calculation for the whole period of the run
    """
    to_process = pu_dfs_to_process(input_dict)
    yearly_ewr_results = process_df_results(to_process)
    
    # aggregate by "gauge","pu","ewrCode"
    final_summary_output = (yearly_ewr_results
    .groupby(["scenario","gauge","pu","ewrCode"])
    .agg( EventYears = ("eventYears", sum),
          Frequency = ("eventYears", get_frequency),
          AchievementCount = ("numAchieved", sum),
          AchievementPerYear = ("numAchieved", 'mean'),
          EventCount = ("numEvents",sum),
          EventsPerYear = ("numEvents",'mean'),
          ThresholdDays = ("totalEventDays", sum),
          InterEventExceedingCount = ("daysBetweenEvents" , sum),
          NoDataDays =  ("missingDays" , sum),
          TotalDays = ("totalPossibleDays" , sum),
          )
    )
    
    # summarize gauge events
    
    events_to_process = get_events_to_process(events)
    ewr_event_stats = process_ewr_events_stats(events_to_process)
    
    # join summary with gauge events
    
    final_summary_output = final_summary_output.merge(ewr_event_stats, 
                                                      'left',
                                                      left_on=['gauge','pu','ewrCode'], 
                                                      right_on=['gauge','pu',"ewrCode"])
    # Join Ewr parameter to summary
    
    EWR_table, bad_EWRs = data_inputs.get_EWR_table()
    
    EWR_table = EWR_table[['gauge','PlanningUnitName','code','frequency','max inter-event']]
    
    final_summary_output = final_summary_output.merge(EWR_table, 
                                                  'left',
                                                  left_on=['gauge','pu','ewrCode'], 
                                                  right_on=['gauge','PlanningUnitName','code'])
    
    
    selected_columns = [ "scenario",'gauge',
     'pu', 
     'ewrCode',
     'EventYears',
     'Frequency',
     'frequency',
     'AchievementCount',
     'AchievementPerYear',
     'EventCount',
     'totalEvents',
     'EventsPerYear',
     'averageEventLength',
     'ThresholdDays',
     'InterEventExceedingCount',
     'max inter-event',
     'NoDataDays',
     'TotalDays']
    
    final_summary_output = final_summary_output[selected_columns]
    
    renamed_columns = ['Scenario','Gauge', 'PlanningUnit', 'EwrCode', 'EventYears', 'Frequency', 'TargetFrequency',
       'AchievementCount', 'AchievementPerYear', 'EventCount', 'totalEvents','EventsPerYear',
       'AverageEventLength', 'ThresholdDays', 'InterEventExceedingCount',
       'MaxInterEventYears', 'NoDataDays', 'TotalDays']
        
    final_summary_output.columns = renamed_columns
    
    return final_summary_output



from typing import Dict, List
import csv
import os
import urllib
import re
from datetime import datetime

import pandas as pd
from tqdm import tqdm

from . import data_inputs, evaluate_EWRs, summarise_results
#----------------------------------- Scenario testing handling functions--------------------------#

def scenario_handler(scenarios, model_format, allowance, climate):
    '''Takes in scenarios, send to functions to clean, calculate ewrs, and
    summarise results, returns a results summary as a dataframe and a 
    dictionary with yearly results.
    TODO: data profiling to check quality of model inputs prior to running
    '''
    
    # Analyse all scenarios for EWRs
    detailed_results = {}
    detailed_events = {}
    for scenario in tqdm(scenarios, position = 0, leave = True, 
                         bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',
                         desc= 'Evaluating scenarios'):
        
        if model_format == 'Bigmod - MDBA':
            data, header = unpack_model_file(scenarios[scenario], 'Dy', 'Field')
            data = build_MDBA_columns(data, header)
            df_clean = cleaner_MDBA(data)
            df_F, df_L = match_MDBA_nodes(df_clean, data_inputs.get_MDBA_codes())

        elif model_format == 'IQQM - NSW 10,000 years':
            df_unpacked = unpack_IQQM_10000yr(scenarios[scenario])
            df_F, df_L = cleaner_IQQM_10000yr(df_unpacked)

        elif model_format == 'Source - NSW (res.csv)':
            data, header = unpack_model_file(scenarios[scenario], 'Date', 'Field')
            data = build_NSW_columns(data, header)
            df_clean = cleaner_NSW(data)
            df_F, df_L = match_NSW_nodes(df_clean, data_inputs.get_NSW_codes())
            
        gauge_results = {}
        gauge_events = {}
        all_locations = df_F.columns.to_list() + df_L.columns.to_list()
        for gauge in all_locations:
            gauge_results[gauge], gauge_events[gauge] = evaluate_EWRs.calc_sorter(df_F, df_L, gauge, allowance, climate)
        detailed_results[scenario] = gauge_results
        detailed_events[scenario] = gauge_events

    summary_results = summarise_results.summarise(detailed_results, detailed_events)
    
    return detailed_results, summary_results

def unpack_IQQM_10000yr(csv_file):
    '''Ingesting scenario file locations with the NSW specific format for 10,000 year flow timeseries
    returns a dictionary of flow dataframes with their associated header data'''
    
    df = pd.read_csv(csv_file, index_col = 'Date')
    siteList = []
    for location in df.columns:
        gauge = extract_gauge_from_string(location)
        siteList.append(gauge)
    # Save over the top of the column headings with the new list containing only the gauges
    df.columns = siteList
    
    return df
    

def unpack_model_file(csv_file, main_key, header_key):
    '''Ingesting scenario file locations with bigmod format, seperates the flow data and header data
    returns a dictionary of flow dataframes with their associated header data'''
    if csv_file[-3:] != 'csv':
        raise ValueError('''Incorrect file type selected, bigmod format requires a csv file.
                         Rerun the program and try again.''')
    
    #--------functions for pulling main data-------#
    
    def mainData_url(url, line,**kwargs):
        '''Get daily data (excluding the header data); remote file upload'''
        response = urllib.request.urlopen(url)
        lines = [l.decode('utf-8') for l in response.readlines()]
        cr = csv.reader(lines)
        pos = 0
        for row in cr:
            if row[0].startswith(line):
                headerVal = pos
            pos = pos + 1
        if main_key == 'Dy':
            df = pd.read_csv(url, header=headerVal, dtype={'Dy':'int', 'Mn': 'int', 'Year': 'int'}, skiprows=range(headerVal+1, headerVal+2))
        elif main_key == 'Date':
            df = pd.read_csv(url, header=headerVal, skiprows=range(headerVal+1, headerVal+2))
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
        
        return df, headerVal
        
    def mainData(file, line,**kwargs):
        '''Get daily data (excluding the header data); local file upload'''
        if os.stat(file).st_size == 0:
            raise ValueError("File is empty")
        with open(file) as csv_file:
            csv_reader = csv.reader(csv_file) #, delimiter=','
            line_count = 0
            for row in csv_reader:
                if row[0].startswith(line):
                    headerVal = line_count
                line_count = line_count + 1
        if main_key == 'Dy':
            df = pd.read_csv(file, header=headerVal, dtype={'Dy':'int', 'Mn': 'int', 'Year': 'int'}, skiprows=range(headerVal+1, headerVal+2))
        elif main_key == 'Date':
            df = pd.read_csv(file, header=headerVal, skiprows=range(headerVal+1, headerVal+2))
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]

        return df, headerVal
        
    #--------Functions for pulling header data---------#
    
    def headerData_url(url, line, endLine, **kwargs):
        '''Get header data for a remote file upload'''
        response = urllib.request.urlopen(url)
        lines = [l.decode('utf-8') for l in response.readlines()]
        cr = csv.reader(lines)
        pos = 0
        for row in cr:
            if row[0].startswith(line):
                headerVal = pos
            pos = pos + 1
        junkRows = headerVal # Junk rows because rows prior to this value will be discarded
        df = pd.read_csv(url, header=headerVal, nrows = (endLine-junkRows-1), dtype={'Site':'str', 'Measurand': 'str', 'Quality': 'str'})
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
        
        return df
        
    def headerData(file, line, endLine, **kwargs):
        '''Get header data for local file upload'''
        if os.stat(file).st_size == 0:
            raise ValueError("File is empty")
            
        with open(file) as csv_file:
            csv_reader = csv.reader(csv_file)
            line_count = 0
            for row in csv_reader:
                if row[0].startswith(line):
                    headerVal = line_count
                    
                    # Then get column length:
                    num_cols = num_cols = list(range(0,len(row),1))
                    
                line_count = line_count + 1
            junkRows = headerVal # Junk rows because rows prior to this value will be discarded
            
        df = pd.read_csv(file, header=headerVal, usecols = num_cols, 
                         nrows = (endLine-junkRows-1), dtype={'Site':'str', 'Measurand': 'str', 'Quality': 'str'},
                        encoding = 'utf-8', sep =',')
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
            
        return df
    
    if 'http' in csv_file:
        mainData_df, endLine = mainData_url(csv_file, main_key, sep=",")
        headerData_df = headerData_url(csv_file, header_key, endLine, sep=",")
    else:
        mainData_df, endLine = mainData(csv_file, main_key, sep=",")
        headerData_df = headerData(csv_file, header_key, endLine, sep=",")
    
    return mainData_df, headerData_df

def build_MDBA_columns(input_data, input_header):
    '''Takes in the header data file, trims it, and then renames the column headings with the full reference code
    returns a the dataframe with updated column headers'''
    # Clean dataframe
    numRows = int(input_header['Field'].iloc[1]) 
    df = input_header.drop([0,1])     
    df = df.astype(str)
    df['Site'] = df['Site'].map(lambda x: x.strip("'"))
    df['Measurand'] = df['Measurand'].map(lambda x: x.strip("'"))
    df['Quality'] = df['Quality'].map(lambda x: x.strip("'"))

    # Construct refs and save to list:
    listOfCols = []
    for i in range(0, numRows):
        colName = str(df['Site'].iloc[i] + '-' + df['Measurand'].iloc[i] + '-' + df['Quality'].iloc[i])
        listOfCols.append(colName)
    dateList = ['Dy', 'Mn', 'Year']
    listOfCols = dateList + listOfCols

    input_data.columns = listOfCols
    
    return input_data

def build_NSW_columns(input_data, input_header):
    '''Takes in the header data file, trims it, and then renames the column headings with the full reference code
    returns a the dataframe with updated column headers'''
    # Extract unique column ID's from the header:
    
    new_cols = input_header['Name'].to_list()
    new_cols = new_cols[2:]
    new_cols.insert(0, 'Date')
    input_data.columns = new_cols
    
    return input_data

def cleaner_MDBA(input_df):
    '''Ingests dataframe, removes junk columns, fixes date,
    returns formatted dataframe'''
    
    cleaned_df = input_df.rename(columns={'Mn': 'Month', 'Dy': 'Day'})
    cleaned_df['Date'] = pd.to_datetime(cleaned_df[['Year', 'Month', 'Day']], format = '%Y-%m-%d')
    cleaned_df = cleaned_df.drop(['Day', 'Month', 'Year'], axis = 1)
    cleaned_df = cleaned_df.set_index('Date')
    
    return cleaned_df

def cleaner_NSW(input_df):
    '''Convert dates to datetime format, save this to the dataframe index'''
    try:
        input_df['Date'] = pd.to_datetime(input_df['Date'], format = '%d/%m/%Y')
    except ValueError:
        print('Attempted and failed to read in dates in format: dd/mm/yyyy, attempting to look for dates in format: yyyy-mm-dd')
        try:
            input_df['Date'] = pd.to_datetime(input_df['Date'], format = '%Y-%m-%d')
        except ValueError:
            raise ValueError('New date format detected. Cannot read in data')
        print('successfully read in data with yyyy-mm-dd formatting')
    input_df = input_df.set_index('Date')
    
    return input_df

def cleaner_IQQM_10000yr(input_df):
    '''Ingests dataframe, removes junk columns, fixes date, allocates gauges to either flow/level'''
    
    cleaned_df = input_df.copy(deep=True)
    
    try:
        date_start = datetime.strptime(cleaned_df.index[0], '%d/%m/%Y')
        date_end = datetime.strptime(cleaned_df.index[-1], '%d/%m/%Y')
    except ValueError:    
        print('Attempted and failed to read in dates in format: dd/mm/yyyy, attempting to look for dates in format: yyyy-mm-dd')
        try:
            date_start = datetime.strptime(cleaned_df.index[0], '%Y-%m-%d')
            date_end = datetime.strptime(cleaned_df.index[-1], '%Y-%m-%d')
        except ValueError:
            raise ValueError('New date format detected. Cannot read in data')
        print('successfully read in data with yyyy-mm-dd formatting')
    
    date_range = pd.period_range(date_start, date_end, freq = 'D')
    cleaned_df['Date'] = date_range
    cleaned_df = cleaned_df.set_index('Date')
    
    # Split gauges into flow and level, allocate to respective dataframe
    flow_gauges = data_inputs.get_gauges('flow gauges')
    level_gauges = data_inputs.get_gauges('level gauges')
    df_flow = pd.DataFrame(index = cleaned_df.index)
    df_level = pd.DataFrame(index = cleaned_df.index)
    for gauge in cleaned_df.columns:
        if gauge in flow_gauges:
            df_flow[gauge] = cleaned_df[gauge].copy(deep=True)
        elif gauge in level_gauges:
            df_level[gauge] = cleaned_df[gauge].copy(deep=True)
    return df_flow, df_level

def extract_gauge_from_string(input_string):
    '''Takes in a string, pulls out the gauge number from this string'''
    found = re.findall(r'\d+\w', input_string)
    if found:
        for i in found:
            if len(i) >= 6:
                gauge = i
                return gauge
    else:
        return None

def match_MDBA_nodes(input_df, model_metadata):
    '''Checks if the source file columns have EWRs available, returns a flow and level dataframe with only 
    the columns with EWRs available. Renames columns to gauges'''

    flow_gauges = data_inputs.get_gauges('flow gauges')
    level_gauges = data_inputs.get_gauges('level gauges')
    measurands = ['1', '35']
    df_flow = pd.DataFrame(index = input_df.index)
    df_level = pd.DataFrame(index = input_df.index)
    for col in input_df.columns:
        col_clean = col.replace(' ', '')
        site = col_clean.split('-')[0]
        measure = col_clean.split('-')[1]
        if ((measure in measurands) and (model_metadata['SITEID'] == site).any()):
            subset = model_metadata.query("SITEID==@site")
            gauge = subset["AWRC"].iloc[0]
            if gauge in flow_gauges and measure == '1':
                df_flow[gauge] = input_df[col]
            elif gauge in level_gauges and measure == '35':
                df_level[gauge] = input_df[col]

    return df_flow, df_level

def match_NSW_nodes(input_df, model_metadata):
    '''Checks if the source file columns have EWRs available, returns a flow and level dataframe with only 
    the columns with EWRs available. Renames columns to gauges'''
    flow_gauges = data_inputs.get_gauges('flow gauges')
    level_gauges = data_inputs.get_gauges('level gauges')
    df_flow = pd.DataFrame(index = input_df.index)
    df_level = pd.DataFrame(index = input_df.index)
    for col in input_df.columns:
        if not (model_metadata[model_metadata['SITEID'].str.contains(col, na=False)]).empty:
            subset = model_metadata.query("SITEID==@col")
            gauge = subset["AWRC"].iloc[0]
            if gauge in flow_gauges:
                df_flow[gauge] = input_df[col]
            elif gauge in level_gauges:
                df_level[gauge] = input_df[col]

    return df_flow, df_level


class ScenarioHandler:
    
    def __init__(self, scenario_files: List[str], model_format:str, allowance:Dict, climate:str):
        self.scenario_files = scenario_files
        self.model_format = model_format
        self.allowance = allowance
        self.climate = climate
        self.yearly_events = None
        self.pu_ewr_statistics = None
        self.summary_results = None

    def _get_file_names(self, loaded_files):

        file_locations = {}
        for file in loaded_files:
            full_name = file.split('/')
            name_exclude_extension = full_name[-1].split('.csv')[0]
            file_locations[str(name_exclude_extension)] = file
            
        return file_locations
        
    def process_scenarios(self):

        scenarios = self._get_file_names(self.scenario_files)

        # Analyse all scenarios for EWRs
        detailed_results = {}
        detailed_events = {}
        for scenario in tqdm(scenarios, position = 0, leave = True, 
                            bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',
                            desc= 'Evaluating scenarios'):
        
            if self.model_format == 'Bigmod - MDBA':
                data, header = unpack_model_file(scenarios[scenario], 'Dy', 'Field')
                data = build_MDBA_columns(data, header)
                df_clean = cleaner_MDBA(data)
                df_F, df_L = match_MDBA_nodes(df_clean, data_inputs.get_MDBA_codes())

            elif self.model_format == 'IQQM - NSW 10,000 years':
                df_unpacked = unpack_IQQM_10000yr(scenarios[scenario])
                df_F, df_L = cleaner_IQQM_10000yr(df_unpacked)

            elif self.model_format == 'Source - NSW (res.csv)':
                data, header = unpack_model_file(scenarios[scenario], 'Date', 'Field')
                data = build_NSW_columns(data, header)
                df_clean = cleaner_NSW(data)
                df_F, df_L = match_NSW_nodes(df_clean, data_inputs.get_NSW_codes())
            
            gauge_results = {}
            gauge_events = {}
            all_locations = df_F.columns.to_list() + df_L.columns.to_list()
            for gauge in all_locations:
                gauge_results[gauge], gauge_events[gauge] = evaluate_EWRs.calc_sorter(df_F, df_L, gauge, self.allowance, self.climate)
            detailed_results[scenario] = gauge_results
            detailed_events[scenario] = gauge_events

            self.pu_ewr_statistics = detailed_results
            self.yearly_events = detailed_events


    def get_all_events(self)-> pd.DataFrame:

        if not self.yearly_events:
            self.process_scenarios()
        
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events = summarise_results.process_all_events_results(events_to_process)
        return all_events

    def get_yearly_ewr_results(self)-> pd.DataFrame:

        if not self.pu_ewr_statistics:
            self.process_scenarios()

        to_process = summarise_results.pu_dfs_to_process(self.pu_ewr_statistics)
        yearly_ewr_results = summarise_results.process_df_results(to_process)
        return yearly_ewr_results

    def get_ewr_results(self) -> pd.DataFrame:
        
        if not self.pu_ewr_statistics:
            self.process_scenarios()

        return summarise_results.summarise(self.pu_ewr_statistics , self.yearly_events)